// This file is included by the aclfs shared object (preload object)
// and by the libaclfs

static int acl_fd=-1;	// Socket to do request

/*
	Wait to receive a file handle
	Return the file handle or -1.
*/
static int aclfs_receivefd()
{
	int ret = -1;
	char bufctl[sizeof(struct cmsghdr)+1*sizeof(int)];
	memset (bufctl,-1,sizeof(bufctl));
	struct iovec iov[1];
	char buf[1000];
	iov[0].iov_base = buf;
	iov[0].iov_len = sizeof(buf);
	struct msghdr h;
	h.msg_name = 0;
	h.msg_namelen = 0;
	h.msg_iov = iov;
	h.msg_iovlen = 1;
	h.msg_control = bufctl;
	h.msg_controllen = sizeof(bufctl);
	struct cmsghdr *c = (struct cmsghdr*) bufctl;
	c->cmsg_len = sizeof(bufctl);
	//c->cmsg_level = SOL_SOCKET;
	//c->cmsg_type = SCM_RIGHTS;
	h.msg_flags = 0;
	int len = recvmsg (acl_fd,&h,0);
	if (len<=0){
		perror ("recvmsg");
	}else{
		#if 0
			fprintf (stderr,"message recu flags %d len %d %d %u\n"
				,h.msg_flags,len
				,h.msg_controllen,sizeof(bufctl));
			fprintf (stderr,"iov.len %d\n",iov[0].iov_len);
			buf[len] = '\0';
			fprintf (stderr,"buf = :%s:\n",buf);
			fprintf (stderr,"msg_len %d msg_level %d msg_type %d\n"
				,c->cmsg_len
				,c->cmsg_level
				,c->cmsg_type);
		#endif
		/*
			On kernel 2.2, we can check c->cmsg_type and c->cmsg_level
			to find out if there is a file handle

			On kernel 2.0, we can only tell if there is
			a descriptor if it is != -1
			if (c->cmsg_level == SOL_SOCKET
				&& c->cmsg_type == SCM_RIGHTS){
		*/
		int *fd = (int*)CMSG_DATA(c);
		//fprintf (stderr,"fd passing %d\n",fd[0]);
		ret = fd[0];
	}
	return ret;
}

/*
	Wait for a confirmation status.
	Return -1 if the original request failed.
*/
static int aclfs_receive()
{
	int ret = -1;
	union {
		char buf[100];
		ACL_REPLY rep;
	}u;
	int len = read (acl_fd,&u,sizeof(u));
	if (len == sizeof(ACL_REPLY)){
		if (u.rep.status == ACL_STATUS_OK){
			ret = 0;
		}else{
			errno = u.rep.errcode;
		}
	}
	return ret;
}
/*
	Wait for some data
	Return -1 if the original request failed.
*/
static int aclfs_receive_data(char *buf, int size)
{
	int ret = -1;
	ACL_REPLY_DATA rep;
	int len = read (acl_fd,&rep,sizeof(rep));
	if (len >= (int)sizeof(ACL_REPLY)){
		if (rep.status == ACL_STATUS_DATA){
			if (size > rep.len) size = rep.len;
			memcpy (buf,rep.data,size);
			ret = size;
		}else{
			errno = rep.errcode;
		}
	}
	return ret;
}

/*
	Wait for a confirmation of a stat request.
	Return -1 if the original request failed.
*/
static int aclfs_receive_stat (struct stat &buf)
{
	int ret = -1;
	struct {
		ACL_REPLY rep;
		struct stat st;
	}u;
	int len = read (acl_fd,&u,sizeof(u));
	memset (&buf,0,sizeof(struct stat));
	// fprintf (stderr,"receive_stat %d %d %d\n",len,sizeof(u),u.rep.status);
	if (len >= (int)sizeof(ACL_REPLY)){
		errno = u.rep.errcode;
		if (u.rep.status == ACL_STATUS_STAT
			&& u.rep.len == sizeof(struct stat)){
			memcpy (&buf,&u.st,sizeof(struct stat));
			ret = 0;
		}
	}
	return ret;
}

static const char *socket_path = ACLFS_SOCKET;

static int aclfs_connect ()
{
	int ret = -1;
	if (acl_fd != -1){
		ret = 0;
	}else{
		int fd =  socket (AF_UNIX,SOCK_STREAM,0);
		if (fd == -1){
			perror ("aclfs: Can't setup the socket");
		}else{
			struct sockaddr_un un;
			un.sun_family = AF_UNIX;
			strcpy (un.sun_path,socket_path);
			int s = connect(fd,(struct sockaddr*)&un,sizeof(un));
			if (s == -1){
				perror ("aclfs: Can't connect to the dav_ufs server");
			}else{
				acl_fd = fd;
				ret = 0;
			}
		}
	}
	return ret;
}

/*
	Send a request and optionally, a file handle
*/
static void aclfs_sendreq (
	ACL_REQUEST &req,
	int len,		// Length of the request, including the header
	int handle)
{
	char bufctl[sizeof(struct cmsghdr)+sizeof(ACL_UCRED)
		+ sizeof(struct cmsghdr) + sizeof(int)];
	struct cmsghdr *c = (struct cmsghdr*)bufctl;
	struct msghdr h;
	h.msg_control = (void*)bufctl;
	h.msg_controllen = CMSG_LEN(sizeof(ACL_UCRED));
	c->cmsg_len = CMSG_LEN(sizeof(ACL_UCRED));
	c->cmsg_level = SOL_SOCKET;
	c->cmsg_type = SCM_CREDENTIALS;
	ACL_UCRED *uc = (ACL_UCRED*)CMSG_DATA(c);
	uc->pid = getpid();
	uc->uid = getuid();
	uc->gid = getgid();

	if (handle != -1){
		c = (struct cmsghdr*) ((char*)c + c->cmsg_len);
		c->cmsg_len = CMSG_LEN(sizeof(int));
		h.msg_controllen += CMSG_LEN(sizeof(int));
		c->cmsg_level = SOL_SOCKET;
		c->cmsg_type = SCM_RIGHTS;
		int *fds = (int*)CMSG_DATA(c);
		fds[0] = handle;
	}

	struct iovec iov[1];
	iov[0].iov_base = &req;
	iov[0].iov_len = len;

	h.msg_name = 0;
	h.msg_namelen = 0;
	h.msg_iov = iov;
	h.msg_iovlen = 1;
	h.msg_flags = 0;
	void (*old)(int) = signal (SIGPIPE,SIG_IGN);
	int ret = sendmsg (acl_fd,&h,0);
	if (ret==-1){
		for (int i=0; i<3; i++){
			acl_fd = -1;
			aclfs_connect();
			ret = sendmsg (acl_fd,&h,0);
			if (ret == -1 && i!=2) sleep(1);
		}
	}
	if (ret == -1){		
		perror ("sendmsg");
	}
	signal (SIGPIPE,old);
}

static void aclfs_sendreq (
	ACL_REQUEST &req,
	int len)		// Length of the request, including the header
{
	aclfs_sendreq (req,len,-1);
}

ACLFS_STATIC int aclfs_open (const char *path, int mode, int perm)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (path,info,req,len)!=-1){
		//fprintf (stderr,"open :%s: %d %d\n",info.realpath,mode,perm);
		req.cmd = ACL_OPEN;
		req.mode = mode;
		req.perm = perm;
		aclfs_sendreq (req,len);
		ret = aclfs_receivefd();
	}
	return ret;
}

ACLFS_STATIC int aclfs_chdir (const char *path)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (path,info,req,len)!=-1){
		// fprintf (stderr,"chdir :%s: :%s: :%s:\n",path,info.virtpath,info.realpath);
		req.cmd = ACL_CHDIR;
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
		if (ret != -1){
			// We have to set the new current dir in $HOME/n
			aclfs_fakechdir(info.virtpath);
		}
	}
	return ret;
}

ACLFS_STATIC DIR * aclfs_opendir (const char *path)
{
	DIR *ret = NULL;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (path,info,req,len)!=-1){
		req.cmd = ACL_OPENDIR;
		aclfs_sendreq (req,len);
		int fd = aclfs_receivefd();
		// fprintf (stderr,"opendir :%s: :%s: -> %d\n",path,req.paths,fd);
		if (fd != -1){
			ret = (DIR*)fd;
		}
	}
	return ret;
}

ACLFS_STATIC struct dirent *aclfs_readdir (DIR *dir)
{
	int fd = (int)dir;
	static struct dirent ent;
	struct dirent *ret = NULL;
	if (read (fd,&ent,sizeof(ent))==sizeof(ent)) ret = &ent;
	return ret;
}


ACLFS_STATIC int aclfs_closedir (DIR *dir)
{
	int fd = (int)dir;
	return close (fd);
}
ACLFS_STATIC void aclfs_rewinddir (DIR *dir)
{
	fprintf (stderr,"aclfs_rewinddir not done\n");
}
ACLFS_STATIC void aclfs_seekdir (DIR *dir, off_t)
{
	fprintf (stderr,"aclfs_seekdir not done\n");
}

ACLFS_STATIC int aclfs_rename (const char *oldp, const char *newp)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO oldinfo,newinfo;
	ACL_REQUEST oldreq,newreq;
	int oldlen,newlen;
	int oldf = aclfs_translate (oldp,oldinfo,oldreq,oldlen);
	int newf = aclfs_translate (newp,newinfo,newreq,newlen);
	// fprintf (stderr,"aclfs_rename %d %d :%s: :%s:\n",oldf,newf,oldinfo.volume,newinfo.volume);
	if (oldf != -1 && newf !=-1
		&& strcmp(oldinfo.volume,newinfo.volume)==0){
		// fprintf (stderr,"aclfs_rename :%s:[%d] :%s:[%d]\n"
		//	,oldreq.paths,oldlen
		//	,newreq.paths,newlen);
		oldreq.cmd = ACL_RENAME;
		strcpy (oldreq.paths+strlen(oldreq.paths)+1,newreq.paths);
		aclfs_sendreq (oldreq,oldlen+strlen(newreq.paths)+1);
		ret = aclfs_receive();
	}else if (oldf != -1 || newf != -1){
		errno = EXDEV;
		ret = -1;
	}
	return ret;
}
ACLFS_STATIC int aclfs_unlink (const char *path)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (path,info,req,len)!=-1){
		// fprintf (stderr,"unlink :%s:\n",info.realpath);
		req.cmd = ACL_UNLINK;
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
	}
	return ret;
}
ACLFS_STATIC int aclfs_rmdir (const char *path)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (path,info,req,len)!=-1){
		// fprintf (stderr,"rmdir :%s: -> :%s:\n",path,info.realpath);
		req.cmd = ACL_RMDIR;
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
	}
	return ret;
}
ACLFS_STATIC int aclfs_mkdir(const char *path, int mode)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (path,info,req,len)!=-1){
		req.cmd = ACL_MKDIR;
		req.mode = mode;
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
	}
	return ret;
}

ACLFS_STATIC int aclfs_stat (int ver, const char *fname, struct stat *buf)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (fname,info,req,len)!=-1){
		// fprintf (stderr,"stat :%s:\n",info.realpath);
		req.cmd = ACL_STAT;
		aclfs_sendreq (req,len);
		ret = aclfs_receive_stat(*buf);
	}
	return ret;
}
ACLFS_STATIC int aclfs_lstat (int ver, const char *fname, struct stat *buf)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (fname,info,req,len)!=-1){
		// fprintf (stderr,"lstat :%s:\n",info.realpath);
		req.cmd = ACL_LSTAT;
		aclfs_sendreq (req,len);
		ret = aclfs_receive_stat(*buf);
	}
	return ret;
}

ACLFS_STATIC int aclfs_chmod (const char *fname, mode_t mode)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (fname,info,req,len)!=-1){
		req.cmd = ACL_CHMOD;
		req.mode = mode;
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
	}
	return ret;
}
ACLFS_STATIC int aclfs_chown (const char *fname, uid_t uid, gid_t gid)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (fname,info,req,len)!=-1){
		req.cmd = ACL_CHOWN;
		req.mode = uid;
		req.perm = gid;
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
	}
	return ret;
}
ACLFS_STATIC int aclfs_lchown (const char *fname, uid_t uid, gid_t gid)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (fname,info,req,len)!=-1){
		req.cmd = ACL_LCHOWN;
		req.mode = uid;
		req.perm = gid;
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
	}
	return ret;
}


ACLFS_STATIC int aclfs_readlink (const char *fname, char *buf, size_t size)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (fname,info,req,len)!=-1){
		req.cmd = ACL_READLINK;
		aclfs_sendreq (req,len);
		ret = aclfs_receive_data(buf,size);
	}
	return ret;
}
#if 0
ACLFS_STATIC int aclfs_utimes (const char *fname, struct timeval *tvp)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (fname,info,req,len)!=-1){
		req.cmd = ACL_UTIMES;
		if (tvp == NULL){
			req.mode = -1;
			req.perm = -1;
		}else{
			req.mode = tvp[0].tv_sec;
			req.perm = tvp[1].tv_sec;
		}
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
	}
	return ret;
}
#endif
ACLFS_STATIC int aclfs_utime (const char *fname, const struct utimbuf *tvp)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (fname,info,req,len)!=-1){
		req.cmd = ACL_UTIMES;
		if (tvp == NULL){
			req.mode = -1;
			req.perm = -1;
		}else{
			req.mode = tvp->actime;
			req.perm = tvp->modtime;
		}
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
	}
	return ret;
}

ACLFS_STATIC int aclfs_access (const char *fname, int type)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (fname,info,req,len)!=-1){
		req.cmd = ACL_ACCESS;
		req.mode = type;
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
		//fprintf (stderr,"access :%s: ret=%d errno=%d\n",fname,ret,errno);
	}
	return ret;
}

ACLFS_STATIC int aclfs_euidaccess (const char *fname, int type)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (fname,info,req,len)!=-1){
		req.cmd = ACL_EUIDACCESS;
		req.mode = type;
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
		//fprintf (stderr,"euidaccess :%s: ret=%d errno=%d\n",fname,ret,errno);
	}
	return ret;
}

ACLFS_STATIC int aclfs_link (const char *oldp, const char *newp)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO oldinfo,newinfo;
	ACL_REQUEST oldreq,newreq;
	int oldlen,newlen;
	int oldf = aclfs_translate (oldp,oldinfo,oldreq,oldlen);
	int newf = aclfs_translate (newp,newinfo,newreq,newlen);
	// fprintf (stderr,"aclfs_link %d %d :%s: :%s:\n",oldf,newf,oldinfo.volume,newinfo.volume);
	if (oldf != -1 && newf !=-1
		&& strcmp(oldinfo.volume,newinfo.volume)==0){
		oldreq.cmd = ACL_LINK;
		strcpy (oldreq.paths+oldlen,newinfo.relpath);
		aclfs_sendreq (oldreq,oldlen+newlen);
		ret = aclfs_receive();
	}else if (oldf != -1 || newf != -1){
		errno = EXDEV;
		ret = -1;
	}
	return ret;
}
ACLFS_STATIC int aclfs_symlink (const char *oldp, const char *newp)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO newinfo;
	ACL_REQUEST newreq;
	int newlen;
	// fprintf (stderr,"aclfs_symlink %d %d :%s: :%s:\n",oldf,newf,oldinfo.volume,newinfo.volume);
	if (aclfs_translate (newp,newinfo,newreq,newlen)!=-1){
		// oldp does not have to lie in the acl filesystem and do
		// not requires validation on our side. To make things
		// simpler on the remote side, we pass the argument
		// using the same order as rename and link.
		newreq.cmd = ACL_SYMLINK;
		char tmp[PATH_MAX];
		strncpy (tmp,newreq.paths,PATH_MAX);
		strcpy (newreq.paths,oldp);
		int oldlen = strlen(oldp)+1;
		strcpy (newreq.paths+oldlen,tmp);
		aclfs_sendreq (newreq,newlen+oldlen);
		ret = aclfs_receive();
	}
	return ret;
}
ACLFS_STATIC int aclfs_fchmod (int, mode_t)
{
	errno = EPERM;
	return -1;
}
ACLFS_STATIC int aclfs_fchown (int, uid_t, gid_t)
{
	errno = EPERM;
	return -1;
}

ACLFS_STATIC int aclfs_execve (const char *filename, char *const[], char *const[])
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (filename,info,req,len)!=-1){
		//ret = -1;
		//errno = EPERM;
	}
	return ret;
}

ACLFS_STATIC int aclfs_bind (int fd, const struct sockaddr *addr, socklen_t len)
{
	int ret = DISPATCH_DONTCARE;
	// fprintf (stderr,"aclfs_bind fd=%d, len=%d\n",fd,len);
	if (addr->sa_family == AF_INET){
		if (aclfs_isroot()){
			aclfs_connect();
			ACL_REQUEST req;
			req.cmd = ACL_BINDINET;
			req.mode = 0;
			req.perm = 0;
			memcpy (req.paths,addr,len);
			aclfs_sendreq (req,len+sizeof(ACL_REQUEST_BASE),fd);
			ret = aclfs_receive();
		}
	}else if (addr->sa_family == AF_UNIX){
		struct sockaddr_un *un = (struct sockaddr_un *)addr;
		ACLFS_INFO info;
		ACL_REQUEST req;
		int len;
		if (aclfs_translate (un->sun_path,info,req,len)!=-1){
			req.cmd = ACL_BINDUNIX;
			req.mode = 0;
			req.perm = 0;
			aclfs_sendreq (req,len,fd);
			ret = aclfs_receive();
		}
		// fprintf (stderr,"bind receive %d\n",ret);
	}
	return ret;
}


