#include "libaclfs.h"
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>

int main (int argc, char *argv[])
{
	for (int i=1; i<argc; i++){
		const char *arg = argv[i];
		if (strcmp(arg,"--auth")==0){
			if (aclfs_setcred (argv[i+1],argv[i+2])==-1){
				fprintf (stderr,"Invalid user/password\n");
				return -1;
			}else{
				printf ("Credentials ok\n");
			}
			i+=2;
		}else if (strcmp(arg,"--dir")==0){
			aclfs_setmount (argv[i+1],argv[i+2]);
			i+=2;
		}else if (strcmp(arg,"--help")==0){
			printf ("--auth user password\n"
				"--dir dir_path volume\n");
			return -1;
		}else{
			fprintf (stderr,"Invalid argument %s\n",arg);
			return -1;
		}
	}
	while (1){
		printf ("> "); fflush (stdout);
		char line[1000];
		if (fgets(line,sizeof(line)-1,stdin)==NULL){
			break;
		}
		char cmd[1000],arg1[1000],arg2[1000];
		int narg=0;
		if (strcmp(line,"help\n")==0){
			printf (
				"quit\n"
				"help\n"
				"auth user password\n"
				"dir directory_path\n"
				"cat file_path\n"
				);
		}else if (strcmp(line,"quit\n")==0){
			break;
		}else if ((narg=sscanf (line,"%s %s %s",cmd,arg1,arg2))<2){
			fprintf (stderr,"Invalid command\n");
		}else if (strcmp(cmd,"auth")==0 && narg == 3){
			int ok = aclfs_setcred (arg1,arg2);
			printf ("ok=%d\n",ok);
		}else if (strcmp(cmd,"dir")==0 && narg == 2){
			DIR *dir = aclfs_opendir (arg1);
			if (dir == NULL){
				fprintf (stderr,"Can't open dir %s\n",arg1);
			}else{
				struct dirent *ent;
				while ((ent=aclfs_readdir(dir))!=NULL){
					printf ("\t%s\n",ent->d_name);
				}
				aclfs_closedir(dir);
			}
		}else if (strcmp(cmd,"cat")==0 && narg == 2){
			FILE *fd = aclfs_fopen (arg1,"r");
			if (fd == NULL){
				fprintf (stderr,"Can't open file %s (%s)\n",arg1,strerror(errno));
			}else{
				char buf[1000];
				while (fgets(buf,sizeof(buf)-1,fd)!=NULL){
					fputs (buf,stdout);
				}
				aclfs_fclose (fd);
			}
		}else if (strcmp(cmd,"create")==0 && narg == 2){
			FILE *fd = aclfs_fopen (arg1,"w");
			if (fd == NULL){
				fprintf (stderr,"Can't open file %s (%s)\n",arg1,strerror(errno));
			}else{
				printf ("File created\n");
				aclfs_fclose (fd);
			}
		}else if (strcmp(cmd,"mkdir")==0 && narg == 2){
			if (aclfs_mkdir (arg1,0700)==-1){
				fprintf (stderr,"Can't create directory %s (%s)\n",arg1,strerror(errno));
			}else{
				printf ("Directory created\n");
			}
		}else{
			fprintf (stderr,"Invalid command\n");
		}
	}
	return 0;
}


